
# PLY template of parser for baby object language.

"""
Here is the syntax that is parsed:

P : Program
CL : CommandList
C : Command
DL : DeclarationList
D : Declaration
EL : ExpressionList
E : Expression
L : LeftHandSide
IL : IdentifierList
I : Identifier
N : Numeral

P ::=  DL  CL

DL ::=  D;*

D ::=  var I = E  |  proc I ( IL ) : CL end

CL ::=  C;*
   that is, a sequence of zero or more commands, separated by  ;
   ( CL ::=  C  |  C ; CL  |  empty  )

C ::=  L = E  |  if E : CL else CL end  |  print E  |  I ( EL )

EL ::=  E,*
   ( EL ::=   E  |  E , EL  | empty  )

E ::=  N  |  ( E1 OP E2 )  |  L  |  & L
  where  OP ::=  +  |  -

L ::=  I  |  * L

N ::=  string of digits

IL ::=  I,*
   that is, a sequence of zero or more identifiers, separated by  ,
   ( IL ::=  I  |  I , IL  | empty  )

I ::=  strings of letters, not including keywords

The output operator trees conform to these SML datatypes:

type Id = string

datatype LVal =
  Var of Id
| Star of LVal

datatype Exp =
  Num of int
| Deref of LVal
| Ampersand of LVal
| Add of Exp * Exp
| Sub of Exp * Exp

and Comm =
  Assign of LVal * Exp
| Print of Exp
| If of Exp * (Comm list) * (Comm list)
| Call of Id * (Exp list)

and Decl =
  VarDecl of Id
| ProcDecl of Id * (Id list) * (Comm list)

type Prog = (Decl list) * (Comm list)
"""

from ply import *
import a5lex

tokens = a5lex.tokens

# set precedences of infix tokens like this:
#precedence = ( #('left', 'TIMES','DIVIDE'),  etc.


######################################################

# Grammar rules and their corresponding operator trees:

# P ::= DL CL
# type Prog = (Decl list) * (Comm list)
def p_Program(p):
    '''Program : DeclarationList CommandList''' 
    p[0] = "(" + p[1] + ",\n" +  p[2] + ")" 


# DL ::= D;*  (that is,   DL ::=  D ; DL | empty   
# Decl list
def p_DeclarationList1(p):
    '''DeclarationList : Declaration ';' DeclarationList'''
    p[0] = p[1] + "\n :: " + p[3]

def p_DeclarationList2(p):
    '''DeclarationList : empty'''
    p[0] = "nil"


# D ::=  var II = E   |  proc II ( IL ) : CL end
# Decl = Var of Id
#      | Proc of Id * (Id list) * (Comm list)
def p_Declaration1(p):
    '''Declaration : var II '=' Expression '''
    p[0] = "VarDecl(" + p[2] + ", " +  p[4] + ")"

def p_Declaration3(p):
    '''Declaration : proc II '(' IdentifierList ')' ':' CommandList end '''
    p[0] = "ProcDecl("  + p[2] + ", " +  p[4] + ", " +  p[7] + ")"


# CL ::= C;*  (that is,   CL ::=  C | C ; CL | empty )
# Comm list

def p_CommandList(p):
    '''CommandList : Command
                   | Command ';' CommandList
                   | empty '''
    if len(p) == 2 and p[1] != None :  # CommandList : Command
       p[0] =  p[1] + "\n :: nil"
    elif len(p) == 4 :  # CommandList :  Command ';' CommandList
       p[0] = p[1] + "\n :: " + p[3]
    else :  # CommandList : empty,  because p[1] == None
       p[0] = "nil"


# C ::=  L = E  |  if E : CL1 else CL2 end  |  print E  |  II ( EL )
# Comm = Assign of LVal * Exp | Print of Exp
#      | While of Exp * (Comm list) | Call of Id * (Exp list)

def p_Command1(c):
    '''Command : LeftHandSide '=' Expression'''
    c[0] = "Assign(" + c[1] + ", " +  c[3] + ")"

def p_Command2(c):
    '''Command : if Expression ':' CommandList else CommandList end '''
    c[0] = "If(" + c[2] +  ", " +  c[4] + ", " + c[6] + ")"

def p_Command3(c):
    '''Command : print Expression'''
    c[0] = "Print(" +  c[2] + ")"

def p_Command4(c):
    '''Command : II '(' ExpressionList ')' '''
    c[0] = "Call("  + c[1] + ", " +  c[3] + ")"


# EL ::= E,*  (that is,   EL ::=  E ETAIL | empty   
#                         ETAIL ::=  , E ETAIL  |  empty    )
# Exp list
def p_ExpressionList1(t):
    '''ExpressionList : Expression EListTail'''
    t[0] = t[1] + " :: " +  t[2]

def p_ExpressionList2(t):
    '''ExpressionList : empty'''
    t[0] =  "nil"

def p_EListTail1(t):
    '''EListTail : ',' Expression EListTail'''
    t[0] = t[2] + " :: " + t[3]

def p_EListTail2(t):
    '''EListTail : empty'''
    t[0] =  "nil"


# IL ::= I,*  (that is,   IL ::=  II ITAIL | empty   
#                         ITAIL ::=  , II ITAIL  |  empty    )
# Id list
def p_IdentifierList1(t):
    '''IdentifierList : II IListTail'''
    t[0] =  t[1] + " :: " + t[2]

def p_IdentifierList2(t):
    '''IdentifierList :  empty'''
    t[0] =  "nil"

def p_IListTail1(t):
    '''IListTail : ',' II IListTail'''
    t[0] = "" + t[2] + " :: "  + t[3]

def p_IListTail2(t):
    '''IListTail : empty'''
    t[0] =  "nil"


# E ::=  N  |  ( E1 + E2 )  |  LeftHandSide  |  & LeftHandSide   
# Exp = Num of string | Deref of LVal  |  Ampersand of LVal 
#     | Add of Exp * Exp | Sub of Exp * Exp
def p_Expression1(e):
    '''Expression : NUM'''
    e[0] = "Num(" + e[1] +")"

def p_Expression2(e):
    '''Expression : '(' Expression Op Expression ')' '''
    e[0] = e[3]  +  e[2] +  ", " +  e[4] + ")"

def p_Expression3(e):
    '''Expression : LeftHandSide'''
    e[0] = "Deref(" +  e[1] + ")"

def p_Expression4(e):
    '''Expression :  '&' LeftHandSide'''
    e[0] = "Ampersand(" +  e[2] + ")"


def p_op1(o):
    '''Op : '+' '''
    o[0] = "Add("

def p_op2(o):
    '''Op : '-' '''
    o[0] = "Sub("

def p_II(i) :
    '''II : ID'''
    i[0] = "\"" + i[1] + "\""

# LeftHandSide  ::=  Id  |  * LeftHandSide
# Lval = Var of II  |  Star of LVal 
def p_LeftHandSide1(e):
    '''LeftHandSide : II'''
    e[0] = "Var(" + e[1] +")"

def p_LeftHandSide2(e):
    '''LeftHandSide : '*' LeftHandSide '''
    e[0] = "Star(" + e[2] +  ")"


#### Generic empty construction --- no token at all
def p_empty(p):
    '''empty : '''
    pass   # retains  p[0] = None


#### Parse Error handler : prints location of error and quits parser
def p_error(p):
    print "SYNTAX ERROR at LexToken(grammar_phrase, input_word, line_number, char_number):"
    print (15 * " "), p
    print "PARSER QUITS."
    raise Exception

bparser = yacc.yacc()

def parse(data,debug=0):
    bparser.error = 0
    p = bparser.parse(data,debug=debug)
    if bparser.error: return None
    return p

