(********* INTERPRETER FOR MINI-C LANGUAGE 

   The file is organized as follows:
   (1) imports of modules for operator trees,
                              environments (symbol table),
                          and stores (memory vectors)
   (2) interpretation functions for operator trees
   (3) driver code that starts the interpreter
*)


(*** STEP 1: load modules for trees, environments, stores: *)

use "optree.sml";

(*   Loads these definitions:
type Id = string

datatype LVal =
  Var of Id
| Star of LVal

datatype Exp =
  Num of int
| Deref of Id
| Ampersand of LVal
| Add of Exp * Exp
| Sub of Exp * Exp

datatype Comm =
  Assign of LVal * Exp
| Print of Exp
| If of Exp * (Comm list) * (Comm list)
| Call of Id * (Exp list)

datatype Decl =
  VarDecl of Id * Exp
| ProcDecl of Id * (Id list) * (Comm list)

type Prog = (Decl list) * (Comm list)
*)


use "environment.sml";
(*  loads these definitions:
datatype Denotable = 
    Loc of int  
 |  Closure of (Id list)* (Comm list) * ((Id * Denotable) list)

type Env = (Id * Denotable) list

and operations:
lookup : Id * Env -> Denotable
declare : Id * Denotable * Env -> Env
*)


use "store.sml";
(* loads this definition:
type Store = int list

and operations:
allocCell : Store -> int * Store
fetch : int * Store -> int
update : int * int * Store -> Store
revert : int * Store -> Store
*)


(* printState : Env * Store -> ()
   prints the contents of env and store arguments used by interpreter
*)
fun printState(env: Env, mem: Store) =
    (print "Environment = \n";
     print (Env2string env); print "\n";
     print "Store = \n";
     print (Store2string mem); print "\n\n";
     ()  (* return nothong *)
    )
;

(* STEP 2: define interpretation functions: *)
(******************* INTERPRETATION FUNCTIONS 

   All functions have form,
     interpretATree : ATree -> Env -> Store -> Answer
   The arguments are "curried", that is, supplied one at a time,
   e.g.,   interpretATree t env mem
   The body of each  interpretATree  is written as pattern equations, one for
   each form of operator tree --- see the definitions at the top of the file!

*)
(********** EXCEPTIONS (error messages) *)
exception ProcReferenced of string
exception IntCalled of string
exception CallParamError
;

(* interpretLVal : LValue -> Env -> Store -> int (a location number)  *)
fun interpretLVal(Var(id)) env mem =
       (case lookup(id, env) of
                 Loc v =>  v
             |   Closure c => raise (ProcReferenced id)
       )
|   interpretLVal(Star(lv)) env mem = 
       (print "Star not yet implemented\n";
        ~1)
;

(* interpretExp : Exp -> Env -> Store -> int *)
fun interpretExp (Num(n)) env mem = n
|   interpretExp (Deref(lv)) env mem  = 
       let val loc = interpretLVal lv env mem
       in  fetch(loc, mem)
       end
|   interpretExp (Ampersand(lv)) env mem =
       (print "Ampersand not yet implemented\n";
        ~1)
|   interpretExp (Add(e1, e2)) env mem =
       (interpretExp e1 env mem) + (interpretExp e2 env mem)
|   interpretExp (Sub(e1, e2)) env mem =
       (interpretExp e1 env mem) - (interpretExp e2 env mem)
;

(* interpretExpList : Exp list -> Env -> Store -> int list *)
fun interpretExpList [] env mem = []
|   interpretExpList (e::rest) env mem = (interpretExp e env mem) :: (interpretExpList rest env mem)
;
(* Or, you can do it in one line with the  map  operation:  

fun interpretExpList es env mem =  map (fn e => interpretExp e env mem) es

It is coded like this in Python:
def interpretExpList(es, env, mem): [interpretExp(e, env, mem) | e in es]
*)


(* interpretComm : Comm -> Env -> Store -> Store *)
fun interpretComm (Assign(lv, e)) env mem = 
        let val lhs = interpretLVal lv env mem
        in  update(lhs, (interpretExp e env mem), mem)
        end

|   interpretComm (Print(e)) env mem = 
        let val v = interpretExp e env mem
        in (print ((Int.toString v)^"\n");
            printState(env, mem);
            mem   (* return mem unaltered *)
            )
        end
|   interpretComm (If(e, clist1, clist2)) env mem =
        if  not((interpretExp e env mem) = 0)
        then interpretCommList(clist1) env mem
        else interpretCommList(clist2) env mem
|   interpretComm (Call(id, elist)) env mem = 
        (case lookup(id, env) of
            Loc v => raise (IntCalled id)
        |   Closure (params, body, linkToEnv)  =>
               (print "proc call not yet implemented\n";
                mem)
        )

(* interpretCommList : Comm list -> Env -> Store -> Store 
   NOTICE KEYWORD,  and,  BELOW.  This is needed because
   interpretComm calls  interpretCommList  and vice versa!   *)
and interpretCommList [] env mem = mem
|   interpretCommList (c::rest) env mem = 
        interpretCommList rest env (interpretComm c env mem)
;


(* interpretDecl : Decl -> Env -> Store -> (Env * Store)  *)
fun interpretDecl (VarDecl(id, e)) env mem =
        (print "var declaration initialization not yet implemented\n";
        let val initialvalue = interpretExp e env mem
            val (newloc, newmem) = allocCell mem
            val newenv =  declare(id, Loc(newloc), env)          
        in (newenv, newmem)   (* FIX THIS LINE TO UPDATE THE newmem *)
        end 
        )
|   interpretDecl (ProcDecl(id, params, body)) env mem =
        (declare(id, Closure(params, body, env), env),  mem) (* WOW, that was easy! *)
;

(* interpretDeclList : Decl list -> Env -> Store -> (Env * Store) *)
fun interpretDeclList nil env mem = (env, mem)
|   interpretDeclList (d::rest) env mem =
         let val (newenv, newstore) = interpretDecl d env mem
         in  interpretDeclList rest newenv newstore
         end
;
 
(* interpretProg: Prog -> string 
   computes the meaning of the Prog (a list of Decls and a list of Comms),
   prints the final environment and store,
   and returns a string that tells the user how to finish.
*)
fun interpretProg (dlist, clist)  =
              let val (startenv, startstore) = interpretDeclList dlist emptyenv emptymem
                  val endstore = interpretCommList clist startenv startstore
              in  (print "Successful termination!\n";
                   printState(startenv, endstore);
                   "Press  Ctrl z  to finish sml session.\n")
              end
;

(* STEP 3:  EXECUTE DRIVER CODE:  *)

use "tree.txt";                         (* this defines
val tree = ...operator tree constructed by parser...  *)
print "\nExecution starts:\n";
print (interpretProg tree);


