

(***********  OPERATOR TREES  "module"   *************)

(* Definitions of operator trees: *)

type Id = string

datatype LVal =
  Var of Id
| Star of LVal
;

datatype Exp =
  Num of int
| Deref of LVal
| Ampersand of LVal
| Add of Exp * Exp
| Sub of Exp * Exp
;

datatype Comm =
  Assign of LVal * Exp
| Print of Exp
| If of Exp * (Comm list) * (Comm list)
| Call of Id * (Exp list)
;

datatype Decl =
  VarDecl of Id * Exp
| ProcDecl of Id * (Id list) * (Comm list)
;

type Prog = (Decl list) * (Comm list)
;


(* Functions that reformat all forms of operator trees into strings
   for printing.  The functions have the form:
              OpTree2string :  OpTree -> string
   e.g.,  Comm2string c   returns the string representation of Comm tree,  c
 *)

(*  (* not needed --- use  Int.toString n  instead. *)
fun Int2string n = 
      let 
          fun digit2string n = implode[chr(n + ord(#"0"))]

          fun i2s n =
                if n > 0 
                then (i2s (n div 10))^(digit2string(n mod 10))
                else ""

       in if n = 0 then "0"
                   else if n > 0 then i2s n
                                 else "~"^(i2s (~n))
       end
*)

fun LVal2string(Var(id)) = "Var("^id^")"
|   LVal2string(Star(lv)) = "Star("^LVal2string(lv)^")"

fun Prog2string(ds, cs) =
    "( " ^ DeclList2string(ds) ^ ",  " ^ CommList2string(cs) ^ " )"

and DeclList2string ds = 
        let fun dl2s [] = ""
            |   dl2s (d::ds) = " " ^ Decl2string(d) ^ "," ^ dl2s(ds)
        in "[" ^ dl2s(ds) ^ "]"
        end

and Decl2string (VarDecl(x, e)) = "VarDecl(" ^ x ^ ", " ^ Exp2string(e) ^ ")"
|   Decl2string (ProcDecl(x, ps, cs)) = "ProcDecl(" ^ x ^ ", " ^ IList2string(ps) ^ ", " ^ CommList2string(cs) ^ ")"

and IList2string is =
        let fun il2s [] = ""
            |   il2s (i::s) = " " ^ i ^ "," ^ il2s(s)
        in "[" ^ il2s(is) ^ "]"
        end

and CommList2string cs = 
        let fun cl2s [] = ""
            |   cl2s (c::cs) = " " ^ Comm2string(c) ^ "," ^ cl2s(cs)
        in "[" ^ cl2s(cs) ^ "]"
        end

and Comm2string (Assign(x,e)) = "Assign(" ^ LVal2string(x) ^ ", " ^ Exp2string(e) ^ ")"
|   Comm2string (Print(e)) = "Print(" ^ Exp2string(e) ^ ")"
|   Comm2string (If(e,cs1, cs2)) = "If(" ^ Exp2string(e) ^ ", " ^ CommList2string(cs1) ^ ", " ^ CommList2string(cs2)^ ")"
|    Comm2string (Call(x, es)) = "Call(" ^ x ^ ", " ^ ExprList2string(es) ^")"

and Exp2string (Num(n)) = "Num(" ^ Int.toString(n) ^ ")"
|   Exp2string (Deref(x)) = "Deref(" ^ LVal2string(x) ^ ")"
|   Exp2string (Ampersand(x)) = "Ampersand(" ^ LVal2string(x) ^ ")"
|   Exp2string (Add(e1, e2)) = "Add(" ^ Exp2string(e1) ^ ", " ^ Exp2string(e2) ^ ")"
|   Exp2string (Sub(e1, e2)) = "Sub(" ^ Exp2string(e1) ^ ", " ^ Exp2string(e2) ^ ")"

and ExprList2string es =
        let fun el2s [] = ""
            |   el2s (e::s) = " " ^ Exp2string(e) ^ "," ^ el2s(s)
        in "[" ^ el2s(es) ^ "]"
        end

fun StringList2string s =
        let fun sl2s [] = ""
            |   sl2s (e::s) = " " ^ e ^ "," ^ sl2s(s)
        in "[" ^ sl2s(s) ^ "]"
        end
;

